<?php
/**
 * JoinPointCreator.class
 *
 * @package aowp.aspect.joinpoint
 */
class AOWP_JoinPointCreator {

	/**
	 * ASTからジョインポイントを作ります。
	 *
	 * @param AOWP_PHPElement $ast PHPの構文要素。AOWP_PHPElementのサブクラス。
	 * @return AOWP_JoinPoint or null
	 */
	public static function createJoinPoint($ast, $fileName) 
	{
		switch (get_class($ast)) {
		case 'AOWP_PHPRootElement':
			return new AOWP_ScriptExecutionJoinPoint($ast, $fileName);
						
		case 'AOWP_PHPFunctionElement';
			return new AOWP_FunctionExecutionJoinPoint($ast);
			
		case 'AOWP_PHPMethodElement':
			return new AOWP_MethodExecutionJoinPoint($ast);

		case 'AOWP_PHPFunctionCallElement':
			return new AOWP_FunctionCallJoinPoint($ast);
			
		case 'AOWP_PHPFunctionCallWithVariableElement':
			return new AOWP_FunctionCallJoinPoint($ast);
			
		case 'AOWP_PHPObjectPropertyElement':
			if (self::isMethodCall($ast)) {
				return new AOWP_MethodCallJoinPoint($ast);
			} else if (self::isFieldSet($ast)) {
				return new AOWP_FieldSetJoinPoint($ast);
			} else {
				return new AOWP_FieldGetJoinPoint($ast);
			}

		case 'AOWP_PHPFileIncludeStatementElement':
			return new AOWP_ScriptCallJoinPoint($ast);
			
		case 'AOWP_PHPNewExprElement':
			return new AOWP_ConstructorCallJoinPoint($ast);
			
		case 'AOWP_PHPReferenceVariableElement': // Array
			if (self::is_Set($ast)) {
				
				/**
				 * Example :
				 * follows array $peach isnot set join point.
				 * 
				 * 		$momo[$peach['melon']] = 0;
				 * 
				 * But array $momo is set join point.
				 */
				if (!self::inArrayKey($ast)){
					return new AOWP_ArraySetJoinPoint($ast);
				} else {
					return null;
				}
				
			} else {
				return new AOWP_ArrayGetJoinPoint($ast);
			}
			
		case 'AOWP_PHPVariableElement':
			/**
			 * Not array reference variable, and not method call's instance.
			 * follows not variable join point.
			 * Example :
			 * 
			 * 		$val->method();
			 * 
			 * Example :
			 * 
			 * 		$val[] = 'hoge';
			 * 
			 */ 
			if ( !($ast->getParent() instanceof AOWP_PHPReferenceVariableElement)
					&& !($ast->getParent() instanceof AOWP_PHPObjectOperatorElement )
					&& !self::_isVariableOfFunctionCall($ast)
					&& !self::_isStaticMethodCallWithVariable($ast)
					&& self::isDeclaredGlobal($ast) 
				) 
				{
					if (self::is_Set($ast)) { 
						return new AOWP_GlobalVariableSetJoinPoint($ast);
					} else {
						return new AOWP_GlobalVariableGetJoinPoint($ast);
					}	
				}
			
			return null; // Array reference variable or local variable.	 

//		case 'AOWP_PHPClassConstVariableRefElement':
//			return new AOWP_ClassConstVariableRefJoinPoint($ast);
			
//		case 'AOWP_PHPStaticMethodCallWithVariableElement';
//			return new AOWP_StaticMethodCallWithVariableJoinPoint($ast);
		}
		
		return null;
	}
	
	private static function isMethodCall ($ast)
	{
		return AOWP_ASTUtility::isArrayElement($ast->arguments);
	}
	
	private static function isFieldSet ($ast)
	{
		return self::is_Set($ast) && ($ast->getParent()->objectProperties[count($ast->getParent()->objectProperties) -1 ]===$ast);
	}
	
	private static function is_Set ($ast)
	{
		/**
		 * Example :
		 * follows $nene appears in left expr, but not set join point.
		 * 
		 * 		list($nene) = $momo;
		 */
		return AOWP_ASTConverter::isDeclaredAtLeftExpr($ast)
		&& !($ast->getParent() instanceof AOWP_PHPListElement);
	}
	
	private static function inArrayKey ($ast)
	{
		return ($ast->getParent() instanceof AOWP_PHPReferenceVariableElement);
	}
	
	private static function isDeclaredGlobal ($ast)
	{
		$pos = AOWP_ASTConverter::getInsertableSpace($ast);
		$val = self::_getVariableName($ast);
		
		// Check if super global variable.
		if (
				$val==_POST ||
				$val==_GET ||
				$val==_FILES ||
				$val==_SERVER ||
				$val==_ENV ||
				$val==_COOKIE ||
				$val==_SESSION ||
				$val==_REQUEST 
			)
			return true;
			
		// Check if declared at global area.
		if ($pos->getParent() instanceof AOWP_PHPRootElement)
		{
			return true;
		}
		
		// Check if declared as global variable.
		if (AOWP_ASTUtility::isArrayElement($pos->getParent()->innerStatements))
		{
			for ($i=0; $i<count($pos->getParent()->innerStatements); $i++)
			{
				if($pos->getParent()->innerStatements[$i] instanceof AOWP_PHPDefineVariableStatementElement)
				{
					if ($pos->getParent()->innerStatements[$i]->type=='global')
					{
						$variables = $pos->getParent()->innerStatements[$i]->variables;
						for ($j=0; $j<count($variables); $j++)
						{
							$valName = self::_getVariableName($variables[$j]);
							if ($valName==$val)
							{
								return true;
							}
						}
					}
				}
			}
		}
		return false;
	}
	
	private static function _isVariableOfFunctionCall ($ast)
	{
		return $ast->getParent()->getParent() instanceof AOWP_PHPFunctionCallWithVariableElement;
	}
	
	private static function _isStaticMethodCallWithVariable ($ast)
	{
		//? getParent()->getParent()
		return $ast->getParent() instanceof AOWP_PHPStaticMethodCallWithVariableElement;
	}
	
	private static function _getVariableName ($ast)
	{
		return preg_replace('/^(\$)/', '', AOWP_TemplateEngine::scriptToSource($ast));
	}
}

?>